#define UNICODE

#include <windows.h>
#include <ShellAPI.h>
#include <shlobj.h>
#include <tchar.h>
#include <winsvc.h>
#include <winioctl.h>
#include "resource.h"

BOOL InstallDriver(IN SC_HANDLE, IN LPCTSTR, IN LPCTSTR);
BOOL StartDriver(IN SC_HANDLE, IN LPCTSTR);
BOOL OpenDevice(IN LPCTSTR, HANDLE *);
BOOL StopDriver(IN SC_HANDLE, IN LPCTSTR);
BOOL RemoveDriver(IN SC_HANDLE, IN LPCTSTR);
BOOL UnloadDeviceDriver(const TCHAR *);
BOOL LoadDeviceDriver(const TCHAR *, const TCHAR *, HANDLE *, PDWORD);
BOOL ServiceUnload(PWORD, SC_HANDLE);
BOOL UnloadDeviceDriverAndRemoveService(const TCHAR *);
BOOL RemoveService(const TCHAR *);

#define FILE_DEVICE_NODEL 0x8000

#define CODE_NODEL_INFO	CTL_CODE(FILE_DEVICE_NODEL, 0x800, \
	METHOD_BUFFERED, FILE_ANY_ACCESS)

HANDLE hDevice = INVALID_HANDLE_VALUE;
WCHAR RealPath[0x1000] = { 0 };

BOOL Start();
BOOL Stop();

LRESULT CALLBACK MainWnd(HWND, UINT, WPARAM, LPARAM);

//
// internals stuff
//

#ifndef UNICODE_STRING
typedef struct _UNICODE_STRING 
{
	USHORT Length;
	USHORT MaximumLength;
	PWSTR  Buffer;
} UNICODE_STRING;
typedef UNICODE_STRING *PUNICODE_STRING;
#endif

#ifndef NTSTATUS
typedef LONG NTSTATUS;
#define NT_SUCCESS(Status) ((NTSTATUS)(Status) >= 0)
#define STATUS_SUCCESS      ((NTSTATUS)0x00000000L)
#endif 

#ifndef STATUS_INFO_LENGTH_MISMATCH
#define STATUS_INFO_LENGTH_MISMATCH   ((NTSTATUS)0xC0000004L)
#endif

#ifndef OBJECT_INFORMATION_CLASS
typedef enum _OBJECT_INFORMATION_CLASS {
	ObjectBasicInformation,
	ObjectNameInformation,
	ObjectTypeInformation,
	ObjectAllTypesInformation,
	ObjectHandleInformation
} OBJECT_INFORMATION_CLASS;
#endif

#ifndef OBJECT_NAME_INFORMATION
typedef struct _OBJECT_NAME_INFORMATION 
{
	UNICODE_STRING ObjectName;

} OBJECT_NAME_INFORMATION, *POBJECT_NAME_INFORMATION;
#endif

#ifndef OBJECT_BASIC_INFORMATION
typedef struct _OBJECT_BASIC_INFORMATION 
{
	ULONG                   Unknown1;
	ACCESS_MASK             DesiredAccess;
	ULONG                   HandleCount;
	ULONG                   ReferenceCount;
	ULONG                   PagedPoolQuota;
	ULONG                   NonPagedPoolQuota;
	BYTE                    Unknown2[32];
} OBJECT_BASIC_INFORMATION, *POBJECT_BASIC_INFORMATION;
#endif

NTSTATUS (NTAPI *pNtQueryObject)(IN HANDLE ObjectHandle,
								 IN OBJECT_INFORMATION_CLASS ObjectInformationClass,
								 OUT PVOID ObjectInformation,
								 IN ULONG ObjectInformationLength,
								 OUT PULONG ReturnLength OPTIONAL);

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
					 LPSTR lpCmdLine, int nCmdShow)
{
	WCHAR path[MAX_PATH] = { 0 };
	BROWSEINFO bi = { 0 };
	bi.lpszTitle = L"Select Folder";

	LPITEMIDLIST pidl = SHBrowseForFolder (&bi);

	if (pidl != 0)
	{
		//
		// get the name of the folder
		//
		
		if (!SHGetPathFromIDList(pidl, path))
		{
			MessageBox(0, L"Invalid Folder", L"Error", 
				MB_ICONERROR);
			return 0;
		}

		//
		// free memory
		//

		IMalloc * imalloc = 0;
		if (SUCCEEDED(SHGetMalloc(&imalloc)))
		{
			imalloc->Free(pidl);
			imalloc->Release();
		}

		//
		// check folder
		//

		if (path[1] != (WCHAR) ':')
		{
			MessageBox(0, L"Invalid Folder", L"Error", 
				MB_ICONERROR);
			return 0;
		}
	}

	//
	// get partition real name
	// there are several ways, one is to get the target
	// from the object manager, since the partition name
	// is an object (simbolic link) who's target name 
	// is the real name of the partition we need,
	// i opted for another way, in this way i don't
	// take chances on the location of the object
	// in the object directory, since maybe in 2k it
	// differs from xp, this way i thought of is safer
	// it simply consists in open a handle to the drive
	// where the directory is located and then ask the name
	// of the object i opened, a very simple and compatible way
	// as you can see...
	//

	HINSTANCE hNtDll = LoadLibrary(L"ntdll.dll");

	if (hNtDll == NULL)
	{
		MessageBox(0, L"An error occurred", L"Error", 
			MB_ICONERROR);
		return 0;
	}

	pNtQueryObject = (NTSTATUS (NTAPI *)(HANDLE,
		OBJECT_INFORMATION_CLASS, PVOID, ULONG, PULONG))
		GetProcAddress(hNtDll, "NtQueryObject");

	WCHAR DriveName[30];

	wsprintf(DriveName, L"\\\\.\\%C:", path[0]);

	HANDLE hDrive;

	hDrive = CreateFile(DriveName, GENERIC_READ, FILE_SHARE_READ |
		FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

	struct { UNICODE_STRING Name; WCHAR Buffer[MAX_PATH + 1]; } ObjName;

	ZeroMemory(&ObjName, sizeof (ObjName));

	NTSTATUS nt = pNtQueryObject(hDrive, ObjectNameInformation,
		&ObjName, sizeof (ObjName),NULL);

	CloseHandle(hDrive);

	wsprintf(RealPath, L"%s%s", ObjName.Buffer, &path[2]);

	//
	// load driver
	//

	if (Start() == FALSE)
	{
		MessageBox(0, L"Cannot Load Driver", L"Error", MB_ICONERROR);
		return 0;
	}

	DialogBox(hInstance, (LPCTSTR) IDD_MAIN, 
		NULL, (DLGPROC) MainWnd);

	return 0;
}

LRESULT CALLBACK MainWnd(HWND hWnd, UINT uMsg,
						 WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{

	case WM_CLOSE:
		{
			EndDialog(hWnd, FALSE);

			if (hDevice != INVALID_HANDLE_VALUE)
			{
				Stop();
			}

			break;
		}

	case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{

			case IDC_START:
				{
					if (Start() == FALSE)
					{
						MessageBox(0, L"Cannot Load Driver", L"Error", MB_ICONERROR);
					}
					else
					{
						EnableWindow(GetDlgItem(hWnd, IDC_START), FALSE);
						EnableWindow(GetDlgItem(hWnd, IDC_STOP), TRUE);
					}
					break;
				}

			case IDC_STOP:
				{
					if (Stop() == FALSE)
					{
						MessageBox(0, L"Cannot Unload Driver", L"Error", MB_ICONERROR);
					}
					else
					{
						EnableWindow(GetDlgItem(hWnd, IDC_START), TRUE);
						EnableWindow(GetDlgItem(hWnd, IDC_STOP), FALSE);
					}
					break;
				}
			}

			break;
		}
	}

	return FALSE;
}

BOOL Start()
{
	WCHAR CurDir[MAX_PATH];
	WCHAR Buffer[MAX_PATH];

	//
	// cur dir
	//

	GetModuleFileName(NULL, Buffer, MAX_PATH * sizeof (WCHAR));

	_wsplitpath(Buffer, CurDir, Buffer, NULL, NULL);
	wcscat(CurDir, Buffer);

	//
	// load driver
	//

	if (OpenDevice(L"nodel", &hDevice) == FALSE)
	{
		DWORD Error;

		wsprintf(Buffer, L"%snodel.sys", CurDir);

		if (LoadDeviceDriver(L"nodel", Buffer, &hDevice, &Error) == FALSE)
		{
			return FALSE;
		}
	}
	else
	{
		return FALSE;
	}

	//
	// send path to driver
	//

	DWORD RetBytes;

	if (!DeviceIoControl(hDevice, CODE_NODEL_INFO, 
		RealPath, 0x1000 * sizeof (WCHAR), NULL, 0, &RetBytes, NULL))
	{
		Stop();
		return  FALSE;
	}

	return TRUE;
}

BOOL Stop()
{
	CloseHandle(hDevice);
	UnloadDeviceDriverAndRemoveService(L"nodel");

	hDevice = INVALID_HANDLE_VALUE;

	return TRUE;
}